# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Internationalization helpers."""

from oslo_config import cfg
import oslo_i18n._message as _msg

cfg.CONF.register_opts([
    cfg.BoolOpt('translation_lazy_load', default=True,
                help='Specifies whether to Lazy-Load Translation')
])


def _(msg, domain='pypowervm'):
    """Shortcut method to return a translated message."""
    msg = _msg.Message(msg, domain=domain)
    # If we are supposed to lazy-load, then return the message object
    return msg if cfg.CONF.translation_lazy_load else msg.translate()
